#!/usr/bin/env ruby

$VERBOSE = nil

# settings
login         = { user: 'username', pass: 'password' }
dest_dir      = '/path/to/moved'
frestart      = '/tmp/deluge_manager.restart'
loop_delay    = 10*60 # delay after each run
call_delay    =  0.2  # delay after an RPC call
call_timeout  = 30.0  # wait time before raising a timeout error
timeout_delay = 60    # delay after a timeout error before attempting a retry
sort_time     = 30*60 # force queue auto sort after this numer of seconds

require 'date'   # Date.parse
require 'deluge' # https://github.com/t3hk0d3/deluge-rpc -- successor of deluge-api, gem install deluge-rpc
                 # http://deluge.readthedocs.io/en/develop/core/rpc.html#remote-api

# exit nicely from the main infinite loop
Signal.trap('INT' ){ puts ''; exit }
Signal.trap('TERM'){ puts ''; exit }

# convenience methods
class String
  def tr_id   ; self.split("\n").grep(/ID:/  ).first.to_s.sub('ID: '  ,''); end
  def tr_name ; self.split("\n").grep(/Name:/).first.to_s.sub('Name: ',''); end
end
class Array
  def tr_ids  ; self.map &:tr_id  ; end
  def tr_names; self.map &:tr_name; end
end

# queue auto sort variables
num_torrents = 0 # initial number of files
last_sort = Time.new(2000,1,1) # last sort time

loop do
  # halt daemon once a day (prevent stalling...)
  last_restart = Date.parse(File.read(frestart)) rescue Date.today
  if last_restart != Date.today && Time.now.hour == 2
    puts "halting deluged PID #{`pgrep deluged`.strip}..."
    system "deluge-console halt"
    sleep 60 # let the daemon exit gracefully
    `pkill deluged`
    num_torrents = 0 # force a queue resort
  end

  # restart daemon if not running
  if `pgrep deluged` && $?.to_i != 0
    File.open(frestart, 'w'){|f| f.print Date.today.to_s }
    print "restarting deluged... "
    system "deluged &"
    sleep 60 # let the daemon complete its startup
    puts "new PID #{`pgrep deluged`.strip}"
  end

  begin
    puts "___ #{Time.now.strftime '%F %T'} ___"

    can_auto_resume = !(Time.now.hour > 17 && Time.now.hour < 24)

    dc = Deluge::Rpc::Client.new host: 'localhost', port: 58846,
                             login: login[:user], password: login[:pass],
                             call_timeout: call_timeout
    dc.connect
  
    # puts dc.api_methods.sort
  
    # *** read torrents data (more complete from console)
    # dc.core.get_torrents_status '', ''; nil
    tr_wip, tr_done = `deluge-console info`.
      split(/^.$\n/)[1..-1].to_a. # la prima riga e' vuota
      partition{|i| i =~ /Progress/ }; nil

    # resort the queue after any addition or time interval
    cur_num_torrents = tr_wip.size + tr_done.size
    if num_torrents < cur_num_torrents || (Time.now - last_sort) > sort_time
      sort_time = Time.now
      do_sort_queue = true
    end
    num_torrents = cur_num_torrents

    # *** fix "torrent stuck at 100% Downloading" > https://forum.deluge-torrent.org/viewtopic.php?t=38279
    tr_err, tr_wip = tr_wip.partition{|i| i =~ /Progress: 100.00%/ }; nil
    if (err_ids = tr_err.tr_ids).size > 0
      print "fix: "
      dc.core.force_recheck err_ids             ; sleep call_delay
      puts '#' * err_ids.size
    end

    # *** recheck torrents in error state
    tr_err, tr_wip = tr_wip.partition{|i| i =~ /State: Error/ }; nil
    if (err_ids = tr_err.tr_ids).size > 0
      print "err: "
      dc.core.force_recheck err_ids             ; sleep call_delay
      puts '#' * err_ids.size
    end
  
    # *** pause & force check completed torrents
    tr_tocheck, tr_done = tr_done.partition{|i| i =~ /State: (Seeding|Queued)/ }; nil
    if (tocheck_ids = tr_tocheck.tr_ids).size > 0
      print "chk: "
      dc.core.pause_torrent tocheck_ids         ; sleep call_delay
      dc.core.force_recheck tocheck_ids         ; sleep call_delay
      puts '#' * tocheck_ids.size
    end

    # *** move & remove completed torrents
    tr_tomove, tr_done = tr_done.partition{|i| i =~ /State: Paused/ }; nil
    if (tomove_ids = tr_tomove.tr_ids).size > 0
      print "mov: "
      dc.core.move_storage tomove_ids, dest_dir ; sleep call_delay
      tomove_ids.each do |id|
        dc.core.remove_torrent id, false        ; sleep call_delay
        num_torrents -= 1
        print '#'
      end
      puts ''
    end

    # *** print any unknown completed torrents
    if tr_done.size > 0
      puts "unk: #{tr_done.size} hits:"
      puts tr_done.tr_names.map{|i| "  - #{i}"}
    end

    # *** reorder queue by num seeds
    tr_sorted = tr_wip.map{|info|
      next if info =~ /State: Checking/

      id = info.tr_id

      # resume paused but not completed torrents
      if can_auto_resume && info =~ /State: Pause/ && info !~ /Progress: 100/
        dc.core.resume_torrent [id]; sleep call_delay
      end

      seeds, peers = info.split("\n").grep(/Seeds/).first.gsub('-1','0').
        match(/Seeds: \d+ \((\d+)\) Peers: \d+ \((\d+)\)/).captures.map(&:to_i) rescue [0,0]
      {id: id, seeds: seeds, peers: peers}
    }.compact.sort{|a,b|
      "#{b[:seeds].to_s.rjust(5,'0')}|#{b[:peers].to_s.rjust(5,'0')}" <=>
      "#{a[:seeds].to_s.rjust(5,'0')}|#{a[:peers].to_s.rjust(5,'0')}"
    }; nil

    if !ARGV.include?('-nosort') && do_sort_queue
      # bump first torrents up and the rest to the bottom
      tr_sorted[0...10].to_a.reverse.each_with_index{|tr, i| print "queue top: #{i+1}/10\r";                     dc.core.queue_top    [ tr[:id] ]; sleep call_delay }; puts ''
      tr_sorted[10..-1].to_a.        each_with_index{|tr, i| print "queue btm: #{i+1}/#{tr_sorted.size - 10}\r"; dc.core.queue_bottom [ tr[:id] ]; sleep call_delay }; puts ''
    end

    dc.close

    sleep loop_delay
  rescue Deluge::Rpc::Connection::InvokeTimeoutError
    puts "\r\n--timeout--\n\n"
    dc.close rescue nil
    sleep timeout_delay
  ensure
    puts ''
  end

  ObjectSpace.garbage_collect
end # main loop

# ------------------------------------------------------------------------------
# NOTE: we can also use the 'deluge' gem but with some caveats:
#   0. gem install 'deluge'
#   1. set ctx.ssl_version = :TLSv1 @ $GEM_HOME/gems/lib/deluge/protocol/v3.rb
#   2. require 'deluge' # https://github.com/realmika/deluge-ruby
#   3. add a method to terminate connection
#        module ExposeConPatch; def con; @con; end; end
#        Deluge::Protocol::V3.send(:include, ExposeConPatch)
#        class Deluge
#          #attr_accessor :con
#          def disconnect
#            Thread.list.each{|t| t.exit if t != Thread.current} 
#            @con.con.close
#          end # disconnect
#        end
#   4. use it like this:
#        dc = Deluge.new #'localhost', '58846'
#        dc.login          login[:user], login[:pass]
#        dc.force_recheck  *err_ids
#        dc.resume_torrent *err_ids
#        dc.pause_torrent  *tocheck_ids
#        dc.remove_torrent id, remove_data: false
#        dc.resume_torrent *[id]
#        dc.queue_top      *[ tr[:id] ]
#        dc.queue_bottom   *[ tr[:id] ]
#        dc.disconnect
#   5. manually manage an eventual server shutdown
# ------------------------------------------------------------------------------
